// CMPSCI 383 - Programming Assignment 1
// Justin Aquadro
// 3dcube.h

#ifndef _3DCUBE_H
#define _3DCUBE_H

#include <vector>
#include <list>
#include <string>
#include <sstream>
#include <iostream>

using namespace std;

class C3DCSpace {
public:
    // Constructors
    C3DCSpace ();
    C3DCSpace (int goal);
    C3DCSpace (int goal, int initial);
    ~C3DCSpace ();
    
    // Methods
    string FormatValue ();
    string FormatValuePretty();
    
    // Properties
    int                 goalValue;
    int                 heldValue;
};

class C3DCube {
public:
    // Constructors
    C3DCube ();
    C3DCube (string h);
    ~C3DCube ();
    
    // Methods
    void GenerateFrom (string state);
    void PrintBoard ();
    
    void SwapSpaces (int idx1, int idx2);
    bool ShiftBlank (int xoff, int yoff, int zoff);
    string Signature ();
    
    bool GoalTest ();
    C3DCube* GenerateState (int xoff, int yoff, int zoff);
    list<C3DCube*> GenerateSuccessors ();
    int Cost ();
    int Cost (string h);
    int CostOutOfPlace ();
    int CostManhattan ();
    int CostOutOfPlane ();
    
    int XCoord (int idx); // X = left-to-right
    int YCoord (int idx); // Y = back-to-front
    int ZCoord (int idx); // Z = top-to-bottom
    int CoordToIDX (int x, int y, int z);
    
    // Properties
    vector<C3DCSpace>   spaces;
    int                 blank;
    int                 cost;
    string              signature;
    string              heuristic;
    C3DCube*            parent;
    int                 depth;
    
    bool operator()(const C3DCube* c1, const C3DCube* c2) const
    {
        return ( c1->cost > c2->cost ) ? true : false;
    }
};

//--- Operators ----------------------------------------------------------------

bool operator> (const C3DCube &c1, const C3DCube &c2)
{
    return (c1.cost > c2.cost) ? true : false;
}

bool operator< (const C3DCube &c1, const C3DCube &c2)
{
    return (c1.cost < c2.cost) ? true : false;
}

//--- Constructors -------------------------------------------------------------

C3DCSpace::C3DCSpace (int goal, int initial)
{
    goalValue = goal;
    heldValue = initial;
}

C3DCSpace::C3DCSpace (int goal)
{
    C3DCSpace( goal, goal );
}

C3DCSpace::C3DCSpace ()
{
    C3DCSpace ( 0, 0 );
}

C3DCSpace::~C3DCSpace ()
{
}

//----------

C3DCube::C3DCube (string h)
{
    blank = 0;
    cost = 0;
    signature = "";
    heuristic = h;
    parent = NULL;
    depth = 0;
    spaces.resize( 27 );
}

C3DCube::C3DCube ()
{
    C3DCube( "H0" );
}

C3DCube::~C3DCube ()
{
}

//--- Methods ------------------------------------------------------------------

string C3DCSpace::FormatValue ()
{
    string fv;
    stringstream strstr;
    
    if ( heldValue == 0 )
    {
        fv = "xx";
    }
    else if ( heldValue > 0 && heldValue < 10 )
    {
        strstr << heldValue;
        fv = "0" + strstr.str();
    }
    else
    {
        strstr << heldValue;
        fv = strstr.str();
    }
    
    return fv;
}

string C3DCSpace::FormatValuePretty ()
{
    string fv;
    
    if ( heldValue == 0 )
    {
        fv = "  ";
    }
    else
    {
        fv = FormatValue();
    }
    
    return fv;
}

//----------

void C3DCube::GenerateFrom (string state)
{
    spaces.clear();

    stringstream strstr;
    strstr.str( state );
    
    for ( int x=0; x<27; x++ )
    {
        int idx;
        string pc;
        strstr >> pc;
        
        if ( pc != "xx" )
        {
            stringstream substr;
            substr.str( pc );
            substr >> idx;
            
            spaces[x] = C3DCSpace( x+1, idx );
        }
        else
        {
            blank = x + 1;
            spaces[x] = C3DCSpace( x+1, 0 );
        }
    }
    
    spaces[26].goalValue = 0;
}

void C3DCube::PrintBoard ()
{
    for ( int x=0; x<27; x++ )
    {
        cout << spaces[x].FormatValue() << " ";
    }
    
    cout << "\n";
}

void C3DCube::SwapSpaces (int idx1, int idx2)
{
    int temp;
    
    idx1--;
    idx2--;
    
    temp = spaces[idx1].heldValue;
    spaces[idx1].heldValue = spaces[idx2].heldValue;
    spaces[idx2].heldValue = temp;
    
    signature = "";
}

bool C3DCube::ShiftBlank (int xoff, int yoff, int zoff)
{
    int bcx = XCoord( blank ) + xoff;
    int bcy = YCoord( blank ) + yoff;
    int bcz = ZCoord( blank ) + zoff;
    
    if ( bcx < 1 || bcx > 3 )
    {
        return false;
    }
    if ( bcy < 1 || bcy > 3 )
    {
        return false;
    }
    if ( bcz < 1 || bcz > 3 )
    {
        return false;
    }
    
    int idx = CoordToIDX( bcx, bcy, bcz );
    
    SwapSpaces( blank, idx );
    
    blank = idx;
    
    return true;
}

string C3DCube::Signature ()
{
    if ( signature.size() == 0 )
    {
        for ( int x=0; x<27; x++ )
        {
            if ( spaces[x].heldValue < 10 )
            {
                signature += (char)( '0' + spaces[x].heldValue );
            }
            else
            {
                signature += (char)( 'A' - 10 + spaces[x].heldValue );
            }
        }
    }
    
    return signature;
}

bool C3DCube::GoalTest ()
{
    for ( int x=0; x<27; x++ )
    {
        if ( spaces[x].heldValue != spaces[x].goalValue )
        {
            return false;
        }
    }
    
    return true;
}

C3DCube* C3DCube::GenerateState (int xoff, int yoff, int zoff)
{
    C3DCube* cube = new C3DCube(heuristic);
    cube->depth = depth + 1;
    cube->blank = blank;
    cube->parent = this;
    
    for ( int x=0; x<27; x++ )
    {
        cube->spaces[x] = C3DCSpace( spaces[x].goalValue, spaces[x].heldValue );
    }
    
    if ( !cube->ShiftBlank ( xoff, yoff, zoff ) )
    {
        delete cube;
        return NULL;
    }
    
    cube->Signature();
    cube->Cost();
    
    return cube;
}

list<C3DCube*> C3DCube::GenerateSuccessors ()
{
    list<C3DCube*> slist;

    C3DCube* nc;
    
    if ( ( nc = GenerateState( -1, 0, 0 ) ) != NULL )
    {
        slist.push_back( nc );
    }
    
    if ( ( nc = GenerateState( 1, 0, 0 ) ) != NULL )
    {
        slist.push_back( nc );
    }
    
    if ( ( nc = GenerateState( 0, -1, 0 ) ) != NULL )
    {
        slist.push_back( nc );
    }
    
    if ( ( nc = GenerateState( 0, 1, 0 ) ) != NULL )
    {
        slist.push_back( nc );
    }
    
    if ( ( nc = GenerateState( 0, 0, -1 ) ) != NULL )
    {
        slist.push_back( nc );
    }
    
    if ( ( nc = GenerateState( 0, 0, 1 ) ) != NULL )
    {
        slist.push_back( nc );
    }
    
    return slist;
}

int C3DCube::Cost ()
{
    return Cost( heuristic );
}

int C3DCube::Cost (string h)
{
    if (h == "H0")
    {
        cost = 1;
    }
    else if (h == "H1")
    {
        cost = CostOutOfPlace();
    }
    else if (h == "H2")
    {
        cost = CostManhattan();
    }
    else if (h == "H3")
    {
        cost = CostOutOfPlane();
    }
    else
    {
        cost = -1;
    }
    
    cost += depth;
    
    return cost;
}

int C3DCube::CostOutOfPlace ()
{
    int count = 0;

    for ( int x=0; x<27; x++ )
    {
        if ( spaces[x].heldValue != spaces[x].goalValue &&
             spaces[x].heldValue != 0 )
        {
            count++;
        }
    }

    return count;
}

int C3DCube::CostManhattan ()
{
    int count = 0;
    
    for ( int x=0; x<27; x++ )
    {
        if ( spaces[x].heldValue != 0 )
        {
            int xdiff = abs( XCoord( spaces[x].heldValue ) - XCoord( x + 1 ) );
            int ydiff = abs( YCoord( spaces[x].heldValue ) - YCoord( x + 1 ) );
            int zdiff = abs( ZCoord( spaces[x].heldValue ) - ZCoord( x + 1 ) );
            
            count += xdiff + ydiff + zdiff;
        }
    }

    return count;
}

int C3DCube::CostOutOfPlane ()
{
    int count = 0;
    
    for ( int x=0; x<27; x++ )
    {
        if ( spaces[x].heldValue != 0 )
        {
            if ( XCoord( spaces[x].heldValue ) != XCoord( x + 1 ) )
            {
                count++;
            }
            if ( YCoord( spaces[x].heldValue ) != YCoord( x + 1 ) )
            {
                count++;
            }
            if ( ZCoord( spaces[x].heldValue ) != ZCoord( x + 1 ) )
            {
                count++;
            }
        }
    }
    
    return count;
}

int C3DCube::XCoord (int idx)
{
    idx--;
    return (idx % 3) + 1;
}

int C3DCube::YCoord (int idx)
{
    idx--;
    return ( (idx % 9) / 3 ) + 1;
}

int C3DCube::ZCoord (int idx)
{
    idx--;
    return (idx / 9) + 1;
}

int C3DCube::CoordToIDX (int x, int y, int z)
{
    int idx = 0;
    idx  = (z - 1) * 9;
    idx += (y - 1) * 3;
    idx += x;
    
    return idx;
}

#endif
