// -*- C++ -*-

#ifndef _DYNSTACK_H_
#define _DYNSTACK_H_

#include "dynarray.h"

/**
 * @class DyanmicStack
 * @brief A stack extention of the DynamicArray, grows as needed.
 *
 * Contains the clear() and trim() methods of DyanmicArray, to clear
 * or trim the size of the stack.
 */

template <class ObjType>
class DynamicStack {
public:
	DynamicStack (void)
		: dynArray (),
			internalIndex (0)
		{}
		
	~DynamicStack (void)
	{
		clear();
	}
	
	/// Clears the stack and releases memory from the underlying array.
	inline void clear (void) {
		dynArray.clear();
		internalIndex = 0;
	}
	
	/// Adds a new element to the top of the stack.
	inline ObjType push (ObjType obj) {
		assert (internalIndex >= 0);
	
		dynArray[internalIndex++] = obj;
		return obj;
	}
	
	/// Pops an element off the top of the stack and returns it.
	inline ObjType pop (void) {
		assert (internalIndex > 0);
		return dynArray[--internalIndex];
	}
	
	/// Returns the element on the top of the stack, but leaves it in place.
	inline ObjType peek (void) {
		assert (internalIndex >= 0);
		return dynArray[internalIndex];
	}
	
	/**
	 * Calls the trim function of the underlying DynamicArray, leaving at
	 * most nelts elements on the stack, and possibly shrinking its footprint.
	 */
	inline void trim (int nelts) {
		dynArray.trim (nelts);
	}
	
	/// Returns true if there are no elements left on the stack.
	inline bool empty () {
		return internalIndex <= 0;
	}
	
private:
	
	/// Storage for stack elements.
	DynamicArray<ObjType> dynArray;
	
	/// Internal stack pointer, index into array.
	int internalIndex;
};

#endif
