// -*- C++ -*-

#ifndef _GCCOMBINEHEAP_H_
#define _GCCOMBINEHEAP_H_

template <class SmallHeap,
	  class BigHeap>
class GCCombineHeap : public CombineHeap<SmallHeap, BigHeap> {
	
	typedef CombineHeap<SmallHeap, BigHeap> ParentHeap;
	
	using ParentHeap::_small;
	using ParentHeap::_big;
	
public:

	virtual ~GCCombineHeap (void) {}
	
	/// Clear out the bitmaps prior to marking
  inline void gcZeroBitmaps (void) {
  	_small.gcZeroBitmaps();
  	_big.gcZeroBitmaps();
  }
  
  /// Compare GC and Malloc/Free bitmaps for consistency
	inline void gcBitmapCheck (void) {
  	_small.gcBitmapCheck();
  }
  
  /// Forces a default value into dead memory
  inline void gcBuryDead (void) {
  	_small.gcBuryDead();
  }
  
  inline void gcSweep (void) {
  	_big.gcSweep();
  }
  
  /// Grow the heap(s) if necessary
  inline void gcGrow (size_t rootSize, size_t allocated = 0) {
  	_small.gcGrow (rootSize, allocated);
  }
  
  /// Try to mark the pointer
  inline int markScan (void * ptr) {
  	int mb = _small.markScan (ptr);
  	if (mb == 0) {
  		mb = _big.markScan (ptr);
  	}
  	return mb;
  }
  
  inline void fullScan (void) {
  	_small.fullScan();
  }
  
  inline int gcLiveSize (void) {
  	return _small.gcLiveSize() + _big.gcLiveSize();
  }
  
  inline int gcHeapSize (void) {
  	return _small.gcHeapSize();
  }
  
  inline void * gcResolve (size_t gcclass, size_t gcheap, size_t gcindex) {
  	return _small.gcResolve (gcclass, gcheap, gcindex);
  }
	
};

#endif
