// -*- C++ -*-

#ifndef _GC_LARGEHEAP_H_
#define _GC_LARGEHEAP_H_

#include <assert.h>

#include "checkpoweroftwo.h"
#include "staticlog.h"
#include "mmapwrapper.h"

#include "gccfg.h"
#include "../largeheap.h"

class GCLargeHeap : public LargeHeap {
	
	typedef LargeHeap ParentHeap;
	
public:

	GCLargeHeap ()
		: _gcStack (GCStack::getStack()),
		  _liveSize (0)
		{ }

  void * malloc (size_t sz) {
  	if (GC_VERBOSE) {
	  	char buf[255];
  		sprintf (buf, "*** Allocated From LARGE HEAP\n");
  		fprintf (stderr, buf);
  	}
  	
  	void * ptr = ParentHeap::malloc (sz);
  	gc_set (ptr);
  	
  	_liveSize += get (ptr);

    return ptr;
  }
  
  bool free (void * ptr) {
  	if (GC_USING_FREE && GC_ENFORCING_FREE) {
  		if (gc_get (ptr) ) {
  			gc_clear (ptr);
  		}
  		
  		return ParentHeap::free (ptr);
  	}
  	
  	return true;
  }
  
  void zeroBitmaps () {
  	_objectMark.eraseAll();
  	_liveSize = 0;
  }
  
  void sweep () {
  	// We should delete memory that wasn't found.  But we don't do that yet.
  }
  
  int markScan (void * ptr) {
  	size_t szRemaining = getSize (ptr);
  	
  	if (szRemaining == 0) {
  		return 0;
  	}
  	
  	if (gc_get (ptr) ) {
  		return 1;
  	}
  	
  	gc_set (ptr);
  	
  	_liveSize += get (ptr);
  	
  	// We can't calculate a size difference .. what to do?
  	GC_mse entry;
    entry.address = ptr;
    entry.length = szRemaining;
    
    _gcStack->push (entry);
    
    return 2;
  }
  
  inline int gcLiveSize (void) {
  	return _liveSize;
  }
	
private:
	
	typedef MyHashMap<void *, bool, SourceHeap> markType;
	
	markType _objectMark;
	
	inline bool gc_get (void * ptr) {
    return _objectMark.get (ptr);
  }
  
  inline void gc_set (void * ptr) {
    // Initialize a range with mark bits.
    size_t sz = get (ptr);
    int iterations = (sz + PAGE_SIZE - 1) / PAGE_SIZE;
    for (int i = 0; i < iterations; i++) {
      _objectMark.set ((char *) ptr + i * PAGE_SIZE, true);
    }
  }
  
  inline void gc_clear (void * ptr) {
    size_t sz = get (ptr);
    int iterations = (sz + PAGE_SIZE - 1) / PAGE_SIZE;
    for (int i = 0; i < iterations; i++) {
      _objectMark.erase ((void *) ((char *) ptr + i * PAGE_SIZE));
    }
  }
  
  GCStack* _gcStack;
  int _liveSize;
  
};


#endif
