// -*- C++ -*-

// This is the beginning of threading support, but it is far from complete
// and provides zero support for threads in its current state.

#ifndef _GC_THREAD_
#define _GC_THREAD_

#include <pthread.h>
#include <semaphore.h>
#include <dlfcn.h>

#include "../lock.h"
#include "../mmapwrapper.h"
#include "gccfg.h"
#include "gcplatform.h"

enum { FINISHED = 1 };
enum { DETACHED = 2 };
enum { MAIN_THREAD = 4 };

Lock gclock;

struct thread_stop_info {
    gc_word last_stop_count;	
    ptr_t stack_ptr;
};

struct start_info {
    void * (*start_routine)(void *);
    void * arg;
    gc_word flags;
    sem_t registered;
};

typedef struct GC_Thread_Rep {
	GC_Thread_Rep * next;
	pthread_t id;
	thread_stop_info stop_info;
	short flags;
	short thread_blocked;
	ptr_t stack_end;
	void * status;
} * GC_thread;


	typedef void * (*gcThreadFunctionType) (void *);

  typedef  
  int (*gc_pthread_create_function) (pthread_t *thread,
				  const pthread_attr_t *attr,
				  gcThreadFunctionType start_routine,
				  void *arg);

// class GCThread {
// Ignore classes for now...

  void GC_thread_exit_proc (void * arg) {
  	GC_thread me;
  	
  	gclock.lock();
  	
  	// ...
  	
  	gclock.unlock();
  }

	void * GC_start_routine (void * arg) {
		int dummy;
		start_info * si = (start_info *) arg;
		GC_thread me;
		void * result;

    pthread_t my_pthread = pthread_self();
    if (GC_VERBOSE && GC_VERBOSE_THREADS) {
    	char buf[255];
    	sprintf (buf, "Starting thread 0x%1x\npid = %1d\nsp = 0x%1x\n", 
    		(long) my_pthread, (long) getpid(), (long) &arg);
    	fprintf (stderr, buf);
    }
    
    gclock.lock();
    
    // ALLOCATE me BEFORE YOU USE IT
    
    // GC_in_thread_creation = TRUE;
    // me = GC_new_thread(my_pthread);
    // GC_in_thread_creation = FALSE;
    
    // me->stop_info.stack_ptr = 0;
    // me->flags = si->flags;
    
    // GC_page_size = 256 is a placeholder
    // me->stack_end = (ptr_t)(((gc_word)(&dummy) + (256 - 1)) & ~(256 - 1));
    // me->stop_info.stack_ptr = me->stack_end - 0x10;
    
    if (GC_VERBOSE & GC_VERBOSE_THREADS) {
    	char buf[255];
    	sprintf (buf, "stack_end = 0x%lx\n", &dummy);
    	fprintf (stderr, buf);
    }
    
    gclock.unlock();
    
    void * (*start)(void *);
    void * start_arg;
    
    start = si->start_routine;
    start_arg = si->arg;
    
    if (GC_VERBOSE & GC_VERBOSE_THREADS) {
    	char buf[255];
    	sprintf (buf, "start_routine = 0x%lx\n", start);
    	fprintf (stderr, buf);
    }
    
    // Last action on si.
    // OK to deallocate.
    sem_post (&(si->registered));
    pthread_cleanup_push (GC_thread_exit_proc, 0);

    result = (*start)(start_arg);
    //result = 0;
    
    me->status = result;
    pthread_cleanup_pop (1);
    
    return result;
	}
	
	
	int GC_pthread_create (pthread_t *thread,
		      const pthread_attr_t *attr,
		      void * (*start_routine) (void *),
		      void * arg) {

    gclock.lock();
    start_info * si = (start_info *) MmapWrapper::map (sizeof (start_info) );
    gclock.unlock();
    
    // if (!parallel_initialized) GC_init_parallel(); // What's this?
    if (si == 0) {
    	return ENOMEM;
    }
    
    sem_init (&(si->registered), 0, 0);
    si->start_routine = start_routine;
    si->arg = arg;
    
    gclock.lock();
    
    // if (!GC_thr_initialized) GC_thr_init(); // Looks important
    
    int detachstate;
    if (attr == 0) {
    	detachstate = PTHREAD_CREATE_JOINABLE;
    } else {
    	pthread_attr_getdetachstate (attr, &detachstate);
    }
    
    gc_word my_flags = 0;
    if (detachstate == PTHREAD_CREATE_DETACHED) {
    	my_flags |= DETACHED;
    }
    si->flags = my_flags;
    
    gclock.unlock();
    
    if (GC_VERBOSE && GC_VERBOSE_THREADS) {
    	char buf[255];
    	sprintf (buf, "About to start new thread 0x%X\n", pthread_self() );
    	fprintf (stderr, buf);
    }
    
    static gc_pthread_create_function real_pthread_create = NULL;
    if (real_pthread_create == NULL) {
      real_pthread_create = (gc_pthread_create_function) dlsym (RTLD_NEXT, "pthread_create");
    }
    int result = (*real_pthread_create)(thread, attr, GC_start_routine, si);
    
    if (GC_VERBOSE && GC_VERBOSE_THREADS) {
    	char buf[255];
    	sprintf (buf, "Started thread 0x%X\n", thread);
    	fprintf (stderr, buf);
    }
    
    // Wait until child has been added to the thread table.
    // This also ensures that we hold onto si until the child is done
    // with it.  Thus it doesn't matter whether it is otherwise
    // visible to the collector.
    if (result == 0) {
    	while (sem_wait (&(si->registered)) != 0) {
    		if (errno != EINTR) {
    			char buf[255];
    			sprintf (buf, "sem_wait failed\n");
    			fprintf (stderr, buf);
    			abort ();
    		}
    	}
    }
    
    sem_destroy (&(si->registered));
    
    gclock.lock();
    MmapWrapper::unmap (si, sizeof (start_info) );
    gclock.unlock();
    
    return result;
	}
	
//};

#endif
