// -*- C++ -*-

/**
 * @file   randomnumbergenerator.h
 * @brief  A generic interface to random number generators.
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 * @note   Copyright (C) 2005 by Emery Berger, University of Massachusetts Amherst.
 */

#ifndef _RANDOMNUMBERGENERATOR_H_
#define _RANDOMNUMBERGENERATOR_H_

#define USE_MERSENNE 0

#if USE_MERSENNE
#include "mersenne.h"
#else
#include "marsaglia.h"
#include "mwc.h"
#endif

#include "gc/gccfg.h"

class RandomNumberGenerator {
public:

  RandomNumberGenerator (unsigned int seed1, unsigned int seed2) 
    : mt (seed1, seed2) {
    //  : mt (362436069, 521288629), _last (0)  {
  }

  inline unsigned int next (void) {
  	if (GC_DISABLE_RANDOM) {
  		_last = rand();
  	} else {
	    _last = mt.next();
	  }
	  
    return _last;
  }
  
  inline unsigned int last (void) {
  	return _last;
  }

private:

	unsigned int _last;

#if USE_MERSENNE
  Mersenne mt;
#else
  MWC mt;
  // Marsaglia mt;
#endif
};

#endif
